{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE ViewPatterns #-}
module Engine.ExperimentCache
( fifoStrategy
, unlimitedStrategy
, ExperimentCache
, lookupExperimentResult
, lookupExperimentHValue
, lookupHValueExperiments
, storeExperimentResult
, storeEngineHValue
, newCache
, Strategy
) where

import Control.Applicative
import Control.Arrow
import Control.Category
import Control.Monad
import Data.IntMap (IntMap)
import Plan.ExperimentMap (Map)
import Data.Maybe
import Data.Sequence (Seq, (><), ViewL(EmptyL, (:<)), (<|), (|>))
import Data.Set (Set)
import Engine.EngineHValue
import Engine.ExperimentResult
import Format
import Id
import Plan
import Prelude hiding ((.), mod, id)
import qualified Data.Foldable as F
import qualified Data.IntMap as IntMap
import qualified Plan.ExperimentMap as Map
import qualified Data.Sequence as Seq
import qualified Data.Set as Set
import qualified Util.BiMap as BiMap
import Util.BiMap (BiMap)
import Util.FCLabels


data ExperimentCache
  = ExperimentCache 
  { weakResults   :: !(Map ExperimentResult)
  , strongResults :: !(Map EngineHValue)
  , valueMap      :: !(IntMap (Seq Entry))
  , strategy      :: !Strategy
  }

newtype Strategy 
  = Strategy 
  { update :: Experiment -> Maybe EngineHValue -> ExperimentCache -> ExperimentCache 
  }

data Entry
  = Entry 
  { entryHValue      :: !EngineHValue
  , entryExperiments :: !(Seq Experiment)
  }

data FIFOData
  = FIFOData { fifoQueue :: !(Seq (Experiment, Maybe EngineHValue))
             , fifoLimit :: !Int
             }


instance Eq Entry where
  e1 == e2 = entryHValue e1 == entryHValue e2

mkLabels [''ExperimentCache, ''Entry]

{- The loop:
 -
 - if exp has a known result
 -  then done
 -  else go through sub-exps and try and convert 
 -       them to ids for running
 -
 -       run total experiment
 -       if new experiment has ?s in it
 -        then remember experiment with experiment result
 -        else if new havlue is in the cache already
 -              then remember current experiment has the existing hvalue/id
 -              else store experiment with id and hvalue
 -
 - So operations:
 -}

lookupExperimentResult :: Experiment -> ExperimentCache ->  Maybe ExperimentResult
lookupExperimentResult exp ec 
  = (Ok <$ Map.lookup (WeakExperiment exp) (strongResults ec)) `mplus` hasDirectResult 
  where
    hasDirectResult = (remapExperimentResult (BiMap.inverseMap bimap)) <$> Map.lookup (WeakExperiment canonicalExp) (weakResults ec) 
    (canonicalExp, bimap) = canonicaliseExperiment exp

lookupExperimentHValue :: Experiment -> ExperimentCache -> Maybe EngineHValue
lookupExperimentHValue (WeakExperiment -> exp) ec 
  = Map.lookup exp (strongResults ec)

storeExperimentResult :: Experiment -> ExperimentResult -> ExperimentCache -> ExperimentCache
storeExperimentResult exp er
  = updateCacheStrategy exp' Nothing
  . mod lWeakResults (Map.insert (WeakExperiment exp') er')
  where
    (exp', bimap) = canonicaliseExperiment exp
    er' = remapExperimentResult (BiMap.forwardMap bimap) er

lookupHValueExperiments :: EngineHValue -> ExperimentCache -> Maybe (Seq Experiment)
lookupHValueExperiments hval ec = do
  possibleEntries <- IntMap.lookup (hashEngineHValue hval) (valueMap ec)
  theEntry <- listToMaybe . F.toList . Seq.filter ((==hval) . entryHValue) $ possibleEntries
  return (entryExperiments theEntry)

storeEngineHValue :: Experiment -> EngineHValue -> ExperimentCache -> ExperimentCache
storeEngineHValue exp hValue 
  = updateCacheStrategy exp (Just hValue)
  . mod lStrongResults (Map.insert (WeakExperiment exp) hValue)
  . mod lValueMap (IntMap.alter fn (hashEngineHValue hValue))
  where
    fn :: Maybe (Seq Entry) -> Maybe (Seq Entry)
    fn = Just . maybe (Seq.singleton newEntry) (\eEntries ->
            let (pre, mPost) = Seq.breakl ((==hValue) . entryHValue) eEntries
            in case Seq.viewl mPost of
                EmptyL        -> newEntry <| pre
                entry :< post -> (mod lEntryExperiments (|> exp) entry) <| pre >< post
                )
      where
        newEntry = Entry { entryHValue      = hValue
                         , entryExperiments = Seq.singleton exp
                         }

updateCacheStrategy :: Experiment -> Maybe EngineHValue -> ExperimentCache -> ExperimentCache
updateCacheStrategy exp mHVal ec = (update . strategy $ ec) exp mHVal ec
 
newCache :: Strategy -> ExperimentCache
newCache strategy = ExperimentCache { weakResults   = Map.empty
                                    , strongResults = Map.empty
                                    , valueMap      = IntMap.empty
                                    , strategy      = strategy
                                    }

fifoStrategy :: Int -> Strategy
fifoStrategy amt = Strategy { update = fifoUpdate FIFOData { fifoQueue = Seq.empty
                                                           , fifoLimit = amt
                                                           }
                            }

fifoUpdate :: FIFOData -> Experiment -> Maybe EngineHValue -> ExperimentCache -> ExperimentCache
fifoUpdate fifo exp mHVal ec = ec' { strategy = Strategy $ fifoUpdate fifo { fifoQueue = toKeep } }
  where
    (toKeep, toRemove) = Seq.splitAt (fifoLimit fifo) . ((exp, mHVal) <|) . fifoQueue $ fifo
    ec' = F.foldl' (flip removeExperiment) ec toRemove

removeExperiment :: (Experiment, Maybe EngineHValue) -> ExperimentCache -> ExperimentCache
removeExperiment (exp, mHVal) 
  = maybe id (\hval -> mod lValueMap (IntMap.alter (removeExpFromEntries hval) (hashEngineHValue hval))) mHVal
  . mod lStrongResults (Map.delete wExp) 
  . mod lWeakResults   (Map.delete wExp)
  where
    wExp = WeakExperiment exp

    removeExpFromEntries :: EngineHValue -> Maybe (Seq Entry) -> Maybe (Seq Entry)
    removeExpFromEntries _ Nothing 
      = {- TODO - "Precondition failure in removeExpFromEntries - should be an entry: " ++ format exp -} Nothing
    removeExpFromEntries hVal (Just entries)
      | Seq.null entries' = Nothing
      | otherwise         = Just entries'
      where
        Just idx = Seq.findIndexL ((==hVal) . entryHValue)  entries
        (before, Seq.viewl -> e :< after) = Seq.splitAt idx entries

        entries' = before >< (maybe id (<|) (updateEntry e) after)

        updateEntry :: Entry -> Maybe Entry
        updateEntry e
          | Seq.null experiments' = Nothing
          | otherwise             = Just $ e { entryExperiments = experiments' }
          where
            experiments' = Seq.filter (/= exp) (entryExperiments e)

unlimitedStrategy :: Strategy
unlimitedStrategy = Strategy { update = \_ _ ec -> ec }
