{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE BangPatterns #-}
module Engine.Profile
( Profile
, recordProfile
, nullProfile
, emptyProfile
, formatProfile
) where

import Control.Applicative
import Data.IORef
import Data.Map (Map)
import Data.Maybe
import Data.Time.Clock.POSIX
import MonadUtils
import qualified Data.Map as Map
import Data.Ord
import Data.List
import Control.Arrow
import Data.Int

data Profile 
  = NoProfile
  | Profile !(IORef ProfileNode) 

data ProfileNode
  = ProfileNode { timeTaken :: !POSIXTime
                , children  :: !(Map String ProfileNode)
                , ticks     :: !Int64
                }

nullProfile :: Profile
nullProfile = NoProfile

emptyProfile :: IO Profile
emptyProfile = Profile <$> newIORef emptyNode

{-# NOINLINE recordProfile #-}
recordProfile :: MonadIO m => Profile -> String -> m a -> m a
recordProfile NoProfile _ c = c
recordProfile (Profile ref) name c = do
  ProfileNode { timeTaken , children , ticks } <- liftIO $ readIORef ref
  let childNode = fromMaybe emptyNode (Map.lookup name children)
  liftIO $ writeIORef ref $! childNode

  !tStart <- liftIO $ getPOSIXTime 
  !res <- c
  () <- res `seq` return ()
  !tEnd <- liftIO $ getPOSIXTime

  !ProfileNode { timeTaken = childTime, children = childChildren, ticks = childTicks } <- liftIO $ readIORef ref
  let newChild = ProfileNode { timeTaken = childTime + (tEnd - tStart)
                             , children  = childChildren
                             , ticks     = childTicks + 1
                             }

  liftIO $ writeIORef ref $! (ProfileNode { timeTaken = timeTaken
                                          , children  = Map.insert name newChild children
                                          , ticks     = ticks
                                          }
                             )
  return $! res


emptyNode :: ProfileNode
emptyNode = ProfileNode { timeTaken = 0
                        , children = Map.empty
                        , ticks    = 0
                        }


formatProfile :: Profile -> IO String
formatProfile NoProfile = return ""
formatProfile (Profile ref) = do
  pNode <- readIORef ref
  let costGraph = formatProfile' 0 "ROOT" pNode
  let indvTimeProfile = toIndvTimeProfile pNode
  let indvCostTable = formatTable (calculateSumCosts "ROOT" indvTimeProfile)

  return $ unlines [costGraph, indvCostTable]


formatProfile' :: Int -> String -> ProfileNode -> String
formatProfile' n name ProfileNode { timeTaken , children , ticks } 
  = replicate n ' ' ++ "├" ++ nextChar ++ "→" ++ name ++ " = " ++ show timeTaken ++ " ( " ++ show indvTimeTaken ++ " ) " ++ show ticks ++ "\n"
    ++ (concatMap (uncurry (formatProfile' (n+1))) (Map.assocs children))
  where
    nextChar
      | Map.null children = "─"
      | otherwise         = "┬"

    indvTimeTaken = timeTaken - (sum . map Engine.Profile.timeTaken . Map.elems $ children)


calculateSumCosts :: String -> ProfileNode -> Map String (POSIXTime, Int64)
calculateSumCosts myName (ProfileNode { timeTaken , children , ticks } ) = Map.unionsWith (\(a,b) (c,d) -> ((a+c),(b+d))) (myMap:childMaps)
  where
    myMap = Map.singleton myName (timeTaken, ticks)
    childMaps :: [Map String (POSIXTime, Int64)]
    childMaps = map (uncurry calculateSumCosts) . Map.assocs $ children
    


toIndvTimeProfile :: ProfileNode -> ProfileNode
toIndvTimeProfile ProfileNode { timeTaken, children, ticks } = ProfileNode { timeTaken = indvTimeTaken
                                                                           , children  = Map.map toIndvTimeProfile children
                                                                           , ticks     = ticks
                                                                           }
  where
    indvTimeTaken = timeTaken - (sum . map Engine.Profile.timeTaken . Map.elems $ children)

formatTable :: Map String (POSIXTime, Int64) -> String
formatTable = unlines . map (uncurry formatLine) . reverse . sortBy (comparing snd) . Map.assocs
  where
    formatLine :: String -> (POSIXTime, Int64) -> String
    formatLine name (time, ticks) = name ++ " " ++ show time ++ " " ++ show ticks
