{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE NamedFieldPuns #-}
module Plan.ExperimentMap
( Map
, empty
, insert
, lookup
, delete
) where

import DataCon
import Data.IntMap (IntMap)
import Data.Maybe
import Data.Monoid
import Data.Sequence (Seq)
import Id
import Plan.Experiment
import Prelude hiding (lookup)
import qualified Data.IntMap as IntMap
import qualified Data.Map as DMap
import qualified Data.Sequence as Seq
import qualified Util.StrictMaybe as Strict
import qualified Util.UniqMap as UniqMap
import Util
import Util.UniqMap (UniqMap)

newtype Map v = Map (RealMap (Strict.First v))

{- This has weak semantics -}
data RealMap v 
  = Empty
  | Split { idRoots       :: !(UniqMap Id (ArgPath v))
          , destructRoots :: !(RealMap (UniqMap DataCon (MIntMap (ArgPath v))))
          , constantRoots :: !(DMap.Map Constant (ArgPath v))
          }

data ArgPath v
  = Stop
  | Step { stopsHere :: !v
         , carrysOn  :: !(ArgMap (ArgPath v))
         }


data ArgMap v
  = ArgEmpty
  | ArgSplit { argExperiment    :: !(RealMap v)
             , argUnconstrained :: !v
             , argBottom        :: !(MIntMap v)
             , argType          :: !v
             }

empty :: Map v
empty = Map Empty

insert :: WeakExperiment -> v -> Map v -> Map v
insert (WeakExperiment e) v (Map rm) = Map (realInsert e (Strict.First (Strict.Just v)) rm)

lookup :: WeakExperiment -> Map v -> Maybe v
lookup (WeakExperiment e) (Map rm) = Strict.maybe Nothing Just 
                                   . Strict.getFirst 
                                   $ walkMap e rm

delete :: WeakExperiment -> Map v -> Map v
delete (WeakExperiment e) (Map rm) = Map (mapUpdate e (const mempty) rm)

realInsert :: (Monoid v) => Experiment -> v -> RealMap v -> RealMap v
realInsert exp v rm = mkSingletonMap exp v `mappend` rm

mkSingletonMap :: (Monoid v) => Experiment -> v -> RealMap v
mkSingletonMap (Experiment { experimentRoot = IdRoot rid
                           , experimentArgs }) !v
  = Split { idRoots       = UniqMap.singleton rid (mkSingletonArgPath experimentArgs v)
          , destructRoots = mempty
          , constantRoots = mempty
          }
mkSingletonMap (Experiment { experimentRoot = DestructRoot { destructCaseOf
                                                           , destructConstructor
                                                           , destructVarIndex
                                                           , destructResultType
                                                           } 
                           , experimentArgs }) v 
  = Split { idRoots = mempty
          , destructRoots = mkSingletonMap destructCaseOf 
                          . UniqMap.singleton destructConstructor 
                          . MIntMap . IntMap.singleton destructVarIndex 
                          . mkSingletonArgPath experimentArgs 
                          $ v
          , constantRoots = mempty
          }
mkSingletonMap (Experiment { experimentRoot = ConstantRoot c
                           , experimentArgs }) !v
  = Split { idRoots       = mempty
          , destructRoots = mempty
          , constantRoots = DMap.singleton c (mkSingletonArgPath experimentArgs v)
          }


mkSingletonArgPath :: (Monoid v) => Seq ExperimentArg -> v -> ArgPath v
mkSingletonArgPath (Seq.viewl -> Seq.EmptyL)  !v = Step { stopsHere = v
                                                       , carrysOn = ArgEmpty
                                                       }
mkSingletonArgPath (Seq.viewl -> x Seq.:< xs) !v = Step { stopsHere = mempty
                                                       , carrysOn  = mkSingletonArgMap x (mkSingletonArgPath xs v)
                                                       }

mkSingletonArgMap :: (Monoid v) => ExperimentArg -> v -> ArgMap v
mkSingletonArgMap (ArgExperiment e) !v = emptyArgMapSplit { argExperiment = mkSingletonMap e v }
mkSingletonArgMap ArgUnconstrained  !v = emptyArgMapSplit { argUnconstrained = v }
mkSingletonArgMap (ArgBottom _ bid) !v = emptyArgMapSplit { argBottom = MIntMap $ IntMap.singleton (fromEnum bid) v } 
mkSingletonArgMap (ArgType _)       !v = emptyArgMapSplit { argType = v }

emptyArgMapSplit :: (Monoid v) => ArgMap v
emptyArgMapSplit = ArgSplit { argExperiment = mempty
                            , argUnconstrained = mempty
                            , argBottom = mempty
                            , argType = mempty
                            }
  
walkMap :: (Monoid v) => Experiment -> RealMap v -> v
walkMap _ Empty = mempty
walkMap (Experiment { experimentRoot = IdRoot x, experimentArgs }) (Split { idRoots }) 
  = walkArgPath experimentArgs (UniqMap.lookup idRoots x)
walkMap (Experiment { experimentRoot = DestructRoot { destructCaseOf
                                                    , destructConstructor
                                                    , destructVarIndex
                                                    } 
                    , experimentArgs }) (Split { destructRoots }) 
  = walkArgPath experimentArgs 
  . fromMaybe mempty . IntMap.lookup destructVarIndex . fromMIntMap
  . flip UniqMap.lookup destructConstructor 
  . walkMap destructCaseOf 
  $ destructRoots
walkMap (Experiment { experimentRoot = ConstantRoot x, experimentArgs }) (Split { constantRoots })
  = walkArgPath experimentArgs (DMap.findWithDefault mempty x constantRoots)

walkArgPath :: (Monoid v) => Seq ExperimentArg -> ArgPath v -> v
walkArgPath _ Stop = mempty
walkArgPath (Seq.viewl -> Seq.EmptyL) (Step { stopsHere }) = stopsHere
walkArgPath (Seq.viewl -> x Seq.:< xs) (Step { carrysOn }) 
  = walkArgPath xs (walkArgMap x carrysOn)

walkArgMap :: (Monoid v) => ExperimentArg -> ArgMap v -> v
walkArgMap _ ArgEmpty = mempty
walkArgMap (ArgExperiment e) (ArgSplit { argExperiment })    = walkMap e argExperiment
walkArgMap ArgUnconstrained  (ArgSplit { argUnconstrained }) = argUnconstrained
walkArgMap (ArgBottom _ bid) (ArgSplit { argBottom })        = fromMaybe mempty . IntMap.lookup (fromEnum bid) . fromMIntMap $ argBottom
walkArgMap (ArgType _)       (ArgSplit { argType })          = argType


mapUpdate :: (Monoid v) => Experiment -> (v -> v) -> RealMap v -> RealMap v
mapUpdate _ f Empty = Empty
mapUpdate (Experiment { experimentRoot = IdRoot x 
                      , experimentArgs }) f s@(Split { idRoots }) 
  = s { idRoots = UniqMap.adjust (argPathUpdate experimentArgs f) x idRoots }
mapUpdate (Experiment { experimentRoot = ConstantRoot x
                      , experimentArgs }) f s@(Split { constantRoots })
  = s { constantRoots = DMap.adjust (argPathUpdate experimentArgs f) x constantRoots }
mapUpdate (Experiment { experimentRoot = DestructRoot { destructCaseOf
                                                      , destructConstructor
                                                      , destructVarIndex}
                      , experimentArgs } ) f s@Split { destructRoots }  
  = s { destructRoots = mapUpdate destructCaseOf (UniqMap.adjust (mIntMapMod $ IntMap.adjust (argPathUpdate experimentArgs f) destructVarIndex) destructConstructor) destructRoots }

argPathUpdate :: (Monoid v) => Seq ExperimentArg -> (v -> v) -> ArgPath v -> ArgPath v
argPathUpdate _ _ Stop = Stop
argPathUpdate (Seq.viewl -> Seq.EmptyL)  f (s@Step { stopsHere }) = s { stopsHere = f stopsHere }
argPathUpdate (Seq.viewl -> x Seq.:< xs) f (s@Step { carrysOn })  = s { carrysOn = argMapUpdate x (argPathUpdate xs f) carrysOn }

argMapUpdate :: (Monoid v) => ExperimentArg -> (v -> v) -> ArgMap v -> ArgMap v
argMapUpdate _ _ ArgEmpty = ArgEmpty
argMapUpdate (ArgExperiment e) f s@ArgSplit { argExperiment }    = s { argExperiment = mapUpdate e f argExperiment }
argMapUpdate ArgUnconstrained f  s@ArgSplit { argUnconstrained } = s { argUnconstrained = f argUnconstrained }
argMapUpdate (ArgBottom _ b) f   s@ArgSplit { argBottom }        = s { argBottom = mIntMapMod (IntMap.adjust f (fromEnum b)) argBottom }
argMapUpdate (ArgType _) f       s@ArgSplit { argType }          = s { argType = f argType }

instance Monoid v => Monoid (RealMap v) where
  mempty = Empty
  Empty `mappend` m2 = m2
  m1 `mappend` Empty = m1
  m1 `mappend` m2 = Split { idRoots       = idRoots m1 `mappend` idRoots m2
                          , destructRoots = destructRoots m1 `mappend` destructRoots m2
                          , constantRoots = DMap.unionWith mappend (constantRoots m1) (constantRoots m2)
                          }

instance Monoid v => Monoid (ArgMap v) where
  mempty = ArgEmpty

  ArgEmpty `mappend` a2 = a2
  a1 `mappend` ArgEmpty = a1
  a1 `mappend` a2       = ArgSplit { argExperiment    = argExperiment a1 `mappend` argExperiment a2
                                   , argUnconstrained = argUnconstrained a1 `mappend` argUnconstrained a2
                                   , argBottom        = argBottom a1 `mappend`  argBottom a2
                                   , argType          = argType a1 `mappend` argType a2
                                   }

instance Monoid v => Monoid (ArgPath v) where
  mempty = Stop

  Stop `mappend` p2 = p2
  p1 `mappend` Stop = p1
  p1 `mappend` p2   = Step { stopsHere = stopsHere p1 `mappend` stopsHere p2
                           , carrysOn  = carrysOn p1 `mappend` carrysOn p2
                           }

