{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE ForeignFunctionInterface #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module Util.GHC where

import Bag
import Control.Applicative
import Control.Monad
import DataCon
import Data.Int
import Data.List
import Data.Maybe
import GHC
import HscTypes
import Id
import IdInfo
import Name
import Outputable
import PrelNames (iNTERACTIVE)
import qualified Control.Monad.Trans as CMT
import qualified Data.Traversable as T
import qualified Linker
import TcRnMonad
import TcType
import Type
import Unify
import Util.FCLabels
import Var
import VarEnv
import VarSet
import TcEnv

mkLabels [''HscEnv, ''DynFlags]

getTyThingId :: TyThing -> Maybe Id
getTyThingId (AnId id)        = Just id
getTyThingId (ADataCon dcon)  = Just (dataConWrapId dcon)
getTyThingId _ = Nothing

initTcForLookup :: HscEnv -> TcM a -> IO a
initTcForLookup hsc_env = liftM (\(msg, mValue) -> fromMaybe (error . show . bagToList . snd $ msg) mValue) . initTc hsc_env HsSrcFile False iNTERACTIVE

instance Show Type where
  show = showPpr

instance Show Name where
  show = showPpr

showDebug ::  (Outputable a) => a -> String
showDebug = showSDocDebug . ppr

listFunTys :: Type -> [Type]
listFunTys (splitFunTys -> (args, res)) = args ++ [res]

idNumArguments :: Id -> Int
idNumArguments = length . listFunTys . idType

instance CMT.MonadIO Ghc where
  liftIO = liftIO

instance MonadThings Ghc where
  
  lookupThing name
    | isWiredInName name = do
      let maybeTyThing = wiredInNameTyThing_maybe name
      return $ fromMaybe (error $ "ho! - lookupThing wired in doesn't work! " ++ showPpr name) maybeTyThing
    | otherwise = do
      withSession $ \hsc -> liftIO $ initTcForLookup hsc (tcLookupGlobal name) -- lookupTypeHscEnv hsc name

instance Ord Type where
  compare = tcCmpType

instance Eq Type where
  (==) = tcEqType

instance Show TvSubst where
  show tvsubst = "TvSubst { " ++ mappings ++ " }"
    where
      inScopeVars = varEnvElts . getInScopeVars . getTvInScope $ tvsubst
      pairs = map (\x -> (x, lookupTyVar tvsubst x)) inScopeVars

      mappings = concat . intersperse ", " . map (\(x,y) -> show x ++ " => " ++ show y) $ pairs

composeTvSubsts :: TvSubst -> TvSubst -> TvSubst
composeTvSubsts l r = mkTvSubst eISS eSE
  where
    lISS = getTvInScope l
    rISS = getTvInScope r
    eISS = extendInScopeSetSet lISS (getInScopeVars rISS)

    lSE = getTvSubstEnv l
    rSE = getTvSubstEnv r
    eSE = composeTvSubst eISS lSE rSE

unionSubsts :: TvSubst -> TvSubst -> TvSubst
unionSubsts l r = mkTvSubst eISS eSE
  where
    lISS = getTvInScope l
    rISS = getTvInScope r
    eISS = extendInScopeSetSet lISS (getInScopeVars rISS)

    lSE = getTvSubstEnv l
    rSE = getTvSubstEnv r
    eSE = lSE `plusVarEnv` rSE

unifyTypes :: Type -> Type -> Maybe TvSubst
unifyTypes target source = {- trace ("unifyTypes: " ++ showDebug target ++ ", " ++ showDebug source ++ " = " ++ showDebug res) -} res
  where
    res = tcUnifyTys instanceBindFun [target] [source]

instanceBindFun :: TyVar -> BindFlag
instanceBindFun tv | isTcTyVar tv && isExistentialTyVar tv = Skolem
                   | otherwise                             = BindMe

foreign import ccall unsafe "getAllocations" getAllocations :: IO Int64

alphaConvertTy :: InScopeSet -> Type -> (InScopeSet, Type, TvSubst)
alphaConvertTy env ty = (env', ty', subst)
  where
    (tyVars, rawType) = splitForAllTys ty
    (env', tyVars') = T.mapAccumL f env tyVars

    subst = zipOpenTvSubst tyVars (mkTyVarTys tyVars')
    rawType' = substTy subst rawType
    ty' = mkForAllTys tyVars' rawType'

    f :: InScopeSet -> TyVar -> (InScopeSet, TyVar)
    f env tv = (env', tv')
      where
        tv' = uniqAway env tv
        env' = env `extendInScopeSet` tv'

alphaConvertRawTy :: InScopeSet -> Type -> Type
alphaConvertRawTy env ty = ty'
  where
    tyVars  = varSetElems . tyVarsOfType $ ty
    tyVars' = snd $ T.mapAccumL f env tyVars
    subst   = zipOpenTvSubst tyVars (mkTyVarTys tyVars')

    ty'    = substTy subst ty

    f :: InScopeSet -> TyVar -> (InScopeSet, TyVar)
    f env tv = (env', tv')
      where
        tv' = uniqAway env tv
        env' = env `extendInScopeSet` tv'

alphaConvert :: InScopeSet -> Id -> (InScopeSet, Id, TvSubst)
alphaConvert env tar = (env', tar', subst)
  where
    (env', ty', subst) = alphaConvertTy env (idType tar)
    tar' = setIdType tar ty'

getAvailableIds :: Module -> Ghc [Id]
getAvailableIds mod = do
  m_mod_info <- getModuleInfo mod
  case m_mod_info of
    Nothing -> return []
    Just modInfo -> do
      let names = {- fromMaybe [] -} (modInfoExports modInfo) -- or modInfoExports or modInfoTyThings ?
      things <- catMaybes <$> mapM lookupGlobalName names
      let allIds  = mapMaybe getTyThingId things
      -- let usableIds = filter (isVanillaIdDetails . idDetails) allIds
      return allIds

getAvailableName_maybe :: GenAvailInfo OccName -> [OccName]
getAvailableName_maybe (Avail name)      = [name]
getAvailableName_maybe (AvailTC _ names) = map mkWorkers names

mkWorkers :: OccName -> OccName
mkWorkers occName
  | isDataOcc occName = mkDataConWorkerOcc occName
  | otherwise = occName


rawGetHValue :: (GhcMonad m) => Id -> m HValue
rawGetHValue id = withSession $ \hscEnv -> do
  rv <- liftIO $ Linker.getHValue hscEnv (idName id)
  return rv
  

getModIface :: ModSummary -> Ghc ModIface
getModIface modsum = do
  withSession $ \hscEnv -> do
    let hpt = hsc_HPT hscEnv
    dflags <- getSessionDynFlags
    let modIface = fromMaybe (error $ "Lookup of non-home packe module: " ++ showPpr modsum)
                             (lookupIfaceByModule dflags hpt emptyPackageIfaceTable (ms_mod modsum))
    return modIface

isVanillaIdDetails :: IdDetails -> Bool
isVanillaIdDetails VanillaId = True
isVanillaIdDetails _         = False

isPropertyId :: Id -> Bool
isPropertyId x = "prop_" `isPrefixOf` nameStr || "prop"  == nameStr
  where
    nameStr = occNameString . getOccName $ x
