{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE ViewPatterns #-}

module Board
( Position(..)
, Piece(..)
, Board
, GameEnd(..)
, emptyBoard
, availablePositions
, allPositions
, placePiece
, nextMove
, firstPiece
, secondPiece
, winningPositions
, lookupPiece
, swapPiece
) where

import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Data.List

import Control.Applicative
import Data.Maybe
import Control.Arrow
import Control.Monad

data Position
  = TL | TM | TR
  | ML | MM | MR
  | BL | BM | BR
  deriving (Eq, Ord, Show, Read, Enum, Bounded)

data Piece 
  = Nought | Cross
  deriving (Eq, Ord, Show)

newtype Board
  = Board 
  { contents  :: (Map Position Piece) 
  }

data GameEnd
  = GameWon Piece
  | GameDrawn
  deriving (Show, Eq, Ord)

emptyBoard :: Board
emptyBoard = Board { contents = Map.empty }

allPositions :: Set Position
allPositions = Set.fromList [minBound .. maxBound]

availablePositions :: Board -> Set Position
availablePositions (Map.keysSet . contents -> usedPositions) 
  = allPositions Set.\\ usedPositions

placePiece :: Board -> Position -> Maybe Board
placePiece board nextPos = do
  (Left nextPiece) <- return $ nextMove board

  guard $ nextPos `Set.member` availablePositions board

  return $ board { contents = Map.insert nextPos nextPiece (contents board) }

lookupPiece :: Position -> Board -> Maybe Piece
lookupPiece pos (contents -> boardContents) = Map.lookup pos boardContents

nextMove :: Board -> Either Piece GameEnd
nextMove board
  | isGameOver                  = Right endGame
  | otherwise                   = Left  nextPiece
  where
    boardContents = contents board
    allPieces = Map.elems boardContents
    (numFirsts, numSeconds) = (length *** length) . partition (== firstPiece) $ allPieces

    winningPiece = msum . map (foldl1' combPieces . map getPiece) $ winningPositions
    isGameOver = (isJust winningPiece) || (numFirsts + numSeconds == 9)
    endGame
      | isNothing winningPiece = GameDrawn
      | (Just p) <- winningPiece = GameWon p

    getPiece :: Position -> Maybe Piece
    getPiece pos = Map.lookup pos boardContents

    combPieces :: Maybe Piece -> Maybe Piece -> Maybe Piece
    combPieces (Just a) (Just b)
      | a == b    = Just a
    combPieces _ _ = Nothing

    nextPiece
      | numFirsts > numSeconds  = secondPiece
      | otherwise               = firstPiece

firstPiece :: Piece
firstPiece = Nought

secondPiece :: Piece
secondPiece = Cross

swapPiece :: Piece -> Piece
swapPiece Nought = Cross
swapPiece Cross  = Nought

winningPositions :: [[Position]]
winningPositions = [ [TL, TM, TR]
                   , [ML, MM, MR]
                   , [BL, BM, BR]
                   , [TL, ML, BL]
                   , [TM, MM, BM]
                   , [TR, MR, BR]
                   , [TL, MM, BR]
                   , [BL, MM, TR]
                   ]
